from galore_torch import GaLoreAdamW, GaLoreAdamW8bit
from peft import LoraConfig, get_peft_model
from torch.optim import AdamW
import json


def load_lora_config(args):
    """Loads LoRa configuration from file"""
    if args.lora_config is not None:
        with open(args.lora_config, "r") as f:
            lora_params = json.load(f)
    else:
        lora_params = {
            "r": args.rank,
            "lora_alpha": args.lora_alpha,
            "lora_dropout": args.lora_dropout,
            "target_modules_finetuning": ["query", "value"],
            "target_modules_pretraining": ["q_proj", "v_proj"]
        }

    target_modules = lora_params["target_modules_finetuning"] if args.mode == "finetuning" else lora_params[
        "target_modules_pretraining"]

    return LoraConfig(
        r=lora_params["r"],
        lora_alpha=lora_params["lora_alpha"],
        lora_dropout=lora_params["lora_dropout"],
        target_modules=target_modules
    )


def load_galore_config(args):
    """Loads GaLore configuration from file"""
    if args.galore_config is not None:
        with open(args.galore_config, "r") as f:
            return json.load(f)
    else:
        return {
            "rank": args.rank,
            "update_proj_gap": args.galore_T,
            "scale": args.galore_alpha,
            "proj_type": "std"
        }


def get_optimizer(args, model):
    """Creates optimizer (GaLore, LoRa, or baseline AdamW)"""
    default_lr = args.lr

    if args.optimizer == "baseline":
        return AdamW(model.parameters(), lr=default_lr, weight_decay=args.weight_decay), model
    elif args.optimizer in ["galore", "galore8bit"]:
        galore_config = load_galore_config(args)
        trainable_params = [p for p in model.parameters() if p.requires_grad and p.dim() > 1]

        param_groups = [
            {"params": trainable_params, **galore_config}
        ]
        optimizer_class = GaLoreAdamW if args.optimizer == "galore" else GaLoreAdamW8bit
        return optimizer_class(param_groups, lr=default_lr, weight_decay=args.weight_decay), model
    elif args.optimizer in ["lora", "lora+galore8bit"]:
        lora_config = load_lora_config(args)
        model = get_peft_model(model, lora_config)
        model.print_trainable_parameters()

        if args.optimizer == "lora":
            return AdamW(model.parameters(), lr=args.lr), model
        else:
            galore_config = load_galore_config(args)
            trainable_params = [p for p in model.parameters() if p.requires_grad and p.dim() > 1]
            param_groups = [
                {"params": trainable_params, **galore_config}
            ]
            return GaLoreAdamW8bit(param_groups, lr=default_lr, weight_decay=args.weight_decay), model
    else:
        raise ValueError(f"Unknown optimizer: {args.optimizer}")