from torch.optim.lr_scheduler import LinearLR

from load_data import load_data
from load_models import get_model
from load_optimizers import get_optimizer
from load_lr_scheduler import get_scheduler

from math import ceil
from logger import init_csv, log_to_csv
from accelerate import Accelerator
import torch
from torch.utils.data import DataLoader
from args import args
import datetime


def train(device, accelerator, scheduler, model, optimizer, dataloader, num_epochs):
    """ training model """
    model, optimizer, dataloader = accelerator.prepare(model, optimizer, dataloader)
    model.train()
    start_time = datetime.datetime.now()
    
    for epoch in range(num_epochs):
        total_loss = 0
        batch_cnt = 0
        for batch in dataloader:
            optimizer.zero_grad()
            
            if args.mode == "pretraining":
                input_ids = batch["input_ids"].to(device)
                attention_mask = batch["attention_mask"].to(device)

                outputs = model(input_ids, attention_mask=attention_mask, labels=input_ids)
            elif args.mode == "finetuning":
                input_ids = batch["input_ids"].to(device)
                attention_mask = batch["attention_mask"].to(device)
                labels = batch["label"].to(device)

                outputs = model(input_ids=input_ids, attention_mask=attention_mask, labels=labels)
            else:
                raise ValueError("Invalid mode. Choose 'pretraining' or 'finetuning'")

            loss = outputs.loss
            accelerator.backward(loss)
            optimizer.step()
            scheduler.step()
            
            compute_time = (datetime.datetime.now() - start_time).total_seconds()
            log_to_csv(epoch + 1, batch_cnt + 1, compute_time, loss.item())

            total_loss += loss.item()
            batch_cnt += 1
        
        avg_loss = total_loss / max(1, batch_cnt)
        print(f"Epoch {epoch+1}, Loss: {avg_loss:.4f}")
    
    return model


if __name__ == "__main__":
    if args.test == "true":
        print("Test mode")
        accelerator = Accelerator()
    else:
        accelerator = Accelerator(mixed_precision="bf16")

    device = "cuda" if torch.cuda.is_available() else "cpu"

    print(f"Running on: {device}")
    print(f"Using optimizer: {args.optimizer}")
    print(f"Arguments: {args}")
    init_csv()

    model, tokenizer = get_model(args)

    dataset = load_data(args, tokenizer)

    shuffle = True if args.shuffle == "true" else False
    if args.mode == "pretraining":
        dataloader = DataLoader(dataset, batch_size=args.batch_size, shuffle=shuffle)
    elif args.mode == "finetuning":
        dataloader = DataLoader(dataset["train"], batch_size=args.batch_size, shuffle=shuffle)
    else:
        raise ValueError("Invalid mode. Choose 'pretraining' or 'finetuning'")

    optimizer, model = get_optimizer(args, model)

    num_batches = len(dataloader) if args.mode == "finetuning" else ceil(args.num_training_tokens / args.batch_size)
    num_steps = args.num_epochs * num_batches
    scheduler = get_scheduler(
        optimizer, args.lr_scheduler, args.warm_up_fraction, num_steps, args.lr, args.lr_min
    )

    trained_model = train(device, accelerator, scheduler, model, optimizer, dataloader, num_epochs=args.num_epochs)

    file_name = f"{args.model}_{args.optimizer}_pretrained" if args.mode == "pretraining" else f"{args.model}_{args.optimizer}_finetuned"
    model_path = f"models/{file_name}"
    trained_model.save_pretrained(model_path)
    tokenizer.save_pretrained(model_path)